/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.apache.mahout.clustering.meanshift;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import org.apache.hadoop.io.Text;
import org.apache.hadoop.io.WritableComparable;
import org.apache.hadoop.mapred.JobConf;
import org.apache.hadoop.mapred.OutputCollector;
import org.apache.mahout.common.distance.DistanceMeasure;
import org.apache.mahout.common.distance.EuclideanDistanceMeasure;
import org.apache.mahout.math.Vector;

public class MeanShiftCanopyClusterer {
  
  private double convergenceDelta = 0;
  // the next canopyId to be allocated
  // private int nextCanopyId = 0;
  // the T1 distance threshold
  private double t1;
  // the T2 distance threshold
  private double t2;
  // the distance measure
  private DistanceMeasure measure;
  
  public MeanShiftCanopyClusterer(JobConf job) {
    configure(job);
  }
  
  public MeanShiftCanopyClusterer(DistanceMeasure aMeasure, double aT1, double aT2, double aDelta) {
    config(aMeasure, aT1, aT2, aDelta);
  }
  
  public double getT1() {
    return t1;
  }
  
  public double getT2() {
    return t2;
  }
  
  /**
   * Configure the Canopy and its distance measure
   * 
   * @param job
   *          the JobConf for this job
   */
  public void configure(JobConf job) {
    try {
      measure = Class.forName(job.get(MeanShiftCanopyConfigKeys.DISTANCE_MEASURE_KEY)).asSubclass(
        DistanceMeasure.class).newInstance();
      measure.configure(job);
    } catch (ClassNotFoundException e) {
      throw new IllegalStateException(e);
    } catch (IllegalAccessException e) {
      throw new IllegalStateException(e);
    } catch (InstantiationException e) {
      throw new IllegalStateException(e);
    }
    // nextCanopyId = 0; // never read?
    t1 = Double.parseDouble(job.get(MeanShiftCanopyConfigKeys.T1_KEY));
    t2 = Double.parseDouble(job.get(MeanShiftCanopyConfigKeys.T2_KEY));
    convergenceDelta = Double.parseDouble(job.get(MeanShiftCanopyConfigKeys.CLUSTER_CONVERGENCE_KEY));
  }
  
  /**
   * Configure the Canopy for unit tests
   * 
   * @param aDelta
   *          the convergence criteria
   */
  public void config(DistanceMeasure aMeasure, double aT1, double aT2, double aDelta) {
    // nextCanopyId = 100; // so canopyIds will sort properly // never read?
    measure = aMeasure;
    t1 = aT1;
    t2 = aT2;
    convergenceDelta = aDelta;
  }
  
  /**
   * Merge the given canopy into the canopies list. If it touches any existing canopy (norm<T1) then add the
   * center of each to the other. If it covers any other canopies (norm<T2), then merge the given canopy with
   * the closest covering canopy. If the given canopy does not cover any other canopies, add it to the
   * canopies list.
   * 
   * @param aCanopy
   *          a MeanShiftCanopy to be merged
   * @param canopies
   *          the List<Canopy> to be appended
   */
  public void mergeCanopy(MeanShiftCanopy aCanopy, List<MeanShiftCanopy> canopies) {
    MeanShiftCanopy closestCoveringCanopy = null;
    double closestNorm = Double.MAX_VALUE;
    for (MeanShiftCanopy canopy : canopies) {
      double norm = measure.distance(canopy.getCenter(), aCanopy.getCenter());
      if (norm < t1) {
        aCanopy.touch(canopy);
      }
      if (norm < t2) {
        if ((closestCoveringCanopy == null) || (norm < closestNorm)) {
          closestNorm = norm;
          closestCoveringCanopy = canopy;
        }
      }
    }
    if (closestCoveringCanopy == null) {
      canopies.add(aCanopy);
    } else {
      closestCoveringCanopy.merge(aCanopy);
    }
  }
  
  /** Emit the new canopy to the collector, keyed by the canopy's Id */
  static void emitCanopy(MeanShiftCanopy canopy, OutputCollector<Text,WritableComparable<?>> collector) throws IOException {
    String identifier = canopy.getIdentifier();
    collector.collect(new Text(identifier), new Text("new " + canopy.toString()));
  }
  
  /**
   * Shift the center to the new centroid of the cluster
   * 
   * @param canopy
   *          the canopy to shift.
   * @return if the cluster is converged
   */
  public boolean shiftToMean(MeanShiftCanopy canopy) {
    Vector centroid = canopy.computeCentroid();
    canopy.setConverged(measure.distance(centroid, canopy.getCenter()) < convergenceDelta);
    canopy.setCenter(centroid);
    canopy.setNumPoints(1);
    canopy.setPointTotal(centroid.clone());
    return canopy.isConverged();
  }
  
  /**
   * Return if the point is covered by this canopy
   * 
   * @param canopy
   *          a canopy.
   * @param point
   *          a Vector point
   * @return if the point is covered
   */
  boolean covers(MeanShiftCanopy canopy, Vector point) {
    return measure.distance(canopy.getCenter(), point) < t1;
  }
  
  /**
   * Return if the point is closely covered by the canopy
   * 
   * @param canopy
   *          a canopy.
   * @param point
   *          a Vector point
   * @return if the point is covered
   */
  public boolean closelyBound(MeanShiftCanopy canopy, Vector point) {
    return measure.distance(canopy.getCenter(), point) < t2;
  }
  
  /**
   * Story: User can exercise the reference implementation to verify that the test datapoints are clustered in
   * a reasonable manner.
   */
  public void testReferenceImplementation() {
    MeanShiftCanopyClusterer clusterer = new MeanShiftCanopyClusterer(new EuclideanDistanceMeasure(), 4.0,
        1.0, 0.5);
    List<MeanShiftCanopy> canopies = new ArrayList<MeanShiftCanopy>();
    // add all points to the canopies
    
    boolean done = false;
    int iter = 1;
    while (!done) {// shift canopies to their centroids
      done = true;
      List<MeanShiftCanopy> migratedCanopies = new ArrayList<MeanShiftCanopy>();
      for (MeanShiftCanopy canopy : canopies) {
        done = clusterer.shiftToMean(canopy) && done;
        clusterer.mergeCanopy(canopy, migratedCanopies);
      }
      canopies = migratedCanopies;
      System.out.println(iter++);
    }
  }
  
  /**
   * This is the reference mean-shift implementation. Given its inputs it iterates over the points and
   * clusters until their centers converge or until the maximum number of iterations is exceeded.
   * 
   * @param points
   *          the input List<Vector> of points
   * @param measure
   *          the DistanceMeasure to use
   * @param numIter
   *          the maximum number of iterations
   */
  public static List<MeanShiftCanopy> clusterPoints(List<Vector> points,
                                                    DistanceMeasure measure,
                                                    double convergenceThreshold,
                                                    double t1,
                                                    double t2,
                                                    int numIter) {
    MeanShiftCanopyClusterer clusterer = new MeanShiftCanopyClusterer(measure, t1, t2, convergenceThreshold);
    
    List<MeanShiftCanopy> canopies = new ArrayList<MeanShiftCanopy>();
    int nextCanopyId = 0;
    for (Vector point : points) {
      clusterer.mergeCanopy(new MeanShiftCanopy(point, nextCanopyId++), canopies);
    }
    
    boolean converged = false;
    for (int iter = 0; !converged && iter < numIter; iter++) {
      converged = runMeanShiftCanopyIteration(canopies, clusterer);
    }
    return canopies;
  }
  
  /**
   * Perform a single iteration over the points and clusters, assigning points to clusters and returning if
   * the iterations are completed.
   * 
   * @param canopies
   *          the List<MeanShiftCanopy> clusters
   */
  public static boolean runMeanShiftCanopyIteration(List<MeanShiftCanopy> canopies,
                                                    MeanShiftCanopyClusterer clusterer) {
    boolean converged = true;
    List<MeanShiftCanopy> migratedCanopies = new ArrayList<MeanShiftCanopy>();
    for (MeanShiftCanopy canopy : canopies) {
      converged = clusterer.shiftToMean(canopy) && converged;
      clusterer.mergeCanopy(canopy, migratedCanopies);
    }
    canopies = migratedCanopies;
    return converged;
    
  }
  
}
